#-----------------------------------------------------------------------------
#
#   Loon - World
#
#-----------------------------------------------------------------------------

from __future__ import division
import random, traceback
from numpy import ndarray, rint, floor, ceil
from pygame import Rect
from albow.resource import get_image, get_sound
from albow.dialogs import alert, ask
from constants import *
from terrain import TerrainCell
from unit_commands import unit_command_list
from unit import population_support_categories
from unit_deposit import add_deposits_to_world
from unit_lander import Lander
from unit_loon import Loon
from unit_rover import Rover
from vector_algebra import vector

#-----------------------------------------------------------------------------

class World(object):
	#  cells             array[,] of Cell
	#  scroll_offset     (float, float)
	#  units             set of Unit
	#  selected_units    set of Unit
	#  units_to_add      set of Unit
	#  units_to_remove   set of Unit
	
	#size = vector(WORLD_XSIZE, WORLD_YSIZE)
	scroll_offset = vector(0, 0) # pixels
	_valid_unit_commands = None

	def __init__(self):
		img = get_image("moon_surface.jpeg")
		self.size = vector(img.get_width() // TILE_WIDTH, img.get_height() // TILE_HEIGHT)
		terrain = ndarray(self.size, object)
		w, h = self.size
		for ix in xrange(w):
			for iy in xrange(h):
				terrain[ix, iy] = TerrainCell()
		self.terrain = terrain
		self.units = set()
		self.selected_units = set()
		self.units_to_add = set()
		self.units_to_remove = set()
		add_deposits_to_world(self)
		self.add_lander()
		self.add_debug_units()
		self.update_stats()
	
	def __getitem__(self, coords):
		try:
			return self.terrain[coords]
		except IndexError:
			pass
	
	def add_lander(self):
		unit = Lander()
		self.add_unit(unit, vector(12, 12))
	
	def add_debug_units(self):
		self.add_unit(Loon(), vector(12, 14))
		self.add_unit(Rover(), vector(14, 16))
	
	def add_unit(self, unit, position):
		unit.position = position
		self.units_to_add.add(unit)
	
	def remove_unit(self, unit):
		self.units_to_remove.add(unit)
	
	def add_new_units(self):
		new = self.units_to_add
		for unit in new:
			if unit.is_fixed:
				for coords in unit.covered_cells():
					cell = self[coords]
					if cell:
						cell.unit = unit
		self.units |= new
		new.clear()
	
	def remove_dead_units(self):
		dead = set(self.units_to_remove)
		for unit in self.units:
			if not unit.is_alive and not unit.leaves_corpse:
				dead.add(unit)
		if dead:
			self.units -= dead
			self.selected_units -= dead
			for unit in dead:
				if unit.is_fixed:
					for coords in unit.covered_cells():
						cell = self[coords]
						if cell and cell.unit is unit:
							cell.unit = None
			for unit in self.units:
				unit.units_removed(dead)
			self.invalidate_unit_commands()
		self.units_to_remove.clear()

	def clear_selected_units(self):
		self.selected_units.clear()
		self.invalidate_unit_commands()
	
	def select_unit(self, unit):
		self.selected_units.add(unit)
		self.invalidate_unit_commands()
	
	def deselect_unit(self, unit):
		self.selected_units.remove(unit)
		self.invalidate_unit_commands()
	
	@property
	def first_selected_unit(self):
		for unit in self.selected_units:
			return unit
	
	@property
	def valid_unit_commands(self):
		#  Returns a list of UnitCommand instances valid for the currently
		#  selected units, in canonical order.
		vc = self._valid_unit_commands
		if vc is None:
			vc = self.find_valid_unit_commands()
			self._valid_unit_commands = vc
		return vc
	
	def invalidate_unit_commands(self):
		self._valid_unit_commands = None
	
	def find_valid_unit_commands(self):
		single = len(self.selected_units) == 1
		vn = set()
		for unit in self.selected_units:
			if not unit.busy:
				vn |= unit.valid_multi_commands
				if single:
					for name in unit.valid_single_commands:
						if unit.command_available(name):
							vn.add(name)
		return [cmd for cmd in unit_command_list
			if cmd.name in vn]
	
	def timestep(self, dt):
		#print "World.timestep:", dt ###
		self.add_new_units()
		self.remove_dead_units()
		self.check_game_over()
		for unit in self.units:
			unit.timestep(dt)
		self.update_stats()
	
	def check_game_over(self):
		pass

	def nearest_unit(self, from_unit, range, filter):
		p0 = from_unit.position
		best_unit = None
		for unit in self.units:
			if filter(unit):
				dsq = magsq(p0 - unit.position)
				if best_unit is None or dsq < best_dsq:
					best_unit = unit
					best_dsq = dsq
		return best_unit

	def site_suitable_for_unit_class(self, world_pt, cls):
		site_size = cls.site_size_required()
		world_min = rint(world_pt - 0.5 * site_size)
		world_max = world_min + site_size
		for unit in self.units:
			if unit.obstructing:
				p = unit.position
				r = unit.radius
				unit_min = rint(p - r)
				unit_max = unit_min + 2 * r
				if ((world_min < unit_max) & (world_max > unit_min)).all():
					return None
		tile = self.tile_at(world_pt)
		if cls.tile_is_suitable(tile):
			return world_min + 0.5 * site_size

	def tile_at(self, p):
		coords = tuple(map(int, floor(p)))
		return self.terrain[coords]

	def item_is_available(self, cls):
		return True

	def update_stats(self):
		pops = {name: 0 for name in population_support_categories}
		for unit in self.units:
			if not unit.under_construction:
				sup = unit.population_support
				if sup:
					cat, pop = sup
					pops[cat] += pop
		self.population_support = pops
