#-----------------------------------------------------------------------------
#
#   Loon - World view
#
#-----------------------------------------------------------------------------

from __future__ import division
from math import floor, ceil
from numpy import floor, ceil, floor_divide, maximum, minimum, rint
from pygame import draw, Color, Rect
from pygame.transform import rotozoom
from pygame.locals import *
from albow.widget import Widget
from albow.resource import get_font, get_image
from albow.utils import frame_rect
from game import game
from constants import *
from vector_algebra import vector, mag

selection_ring = get_image("selection_ring.png")
construction_image = get_image("under_construction.png")

#-----------------------------------------------------------------------------

class WorldView(Widget):
	#  game     Game
	#  client   PlayScreen
	
	#bg_image = get_image("play_bg.png")
	surface_image = get_image("moon_surface.jpeg")
	
	last_click_time = 0
	
	@property
	def world(self):
		return self.game.world
	
	@property
	def first_selected_unit(self):
		for unit in self.selected_units:
			return unit

	def __init__(self, game, client, **kwds):
		Widget.__init__(self, **kwds)
		self.game = game
		self.client = client

	def scroll_offset_in_pixels(self):
		return self.world.scroll_offset
	
	def topleft_cell_coords(self):
		return (0, 0)
	
	def size_in_cells(self):
		w = int(ceil(self.width / TILE_WIDTH))
		h = int(ceil(self.height / TILE_HEIGHT))
		return (w, h)

	def world_to_screen(self, p):
		return rint(p * TILE_SIZE) - self.world.scroll_offset
		#sx, sy = p - self.world.scroll_offset
		#return rint(sx * TILE_WIDTH), rint(sy * TILE_HEIGHT)
	
#	def screen_to_world(self, s, z = 0.0):
#		ox, oy = self.world.scroll_offset
#		sx, sy = s
#		return vector(sx / TILE_WIDTH + ox, (self.height - sy) / TILE_HEIGHT + oy, z)

	def screen_to_world(self, s):
		return (s + self.world.scroll_offset) / TILE_SIZE
		#ox, oy = self.world.scroll_offset
		#sx, sy = s
		#return vector(sx / TILE_WIDTH + ox, sy / TILE_HEIGHT + oy)

	def draw(self, surf):
		world = self.world
		self.draw_background(surf)
		self.draw_tiles(surf, world)
		self.draw_units(surf, world)
		self.client.draw_world_feedback(self, surf)
	
	def draw_background(self, surf):
		x0, y0 = self.scroll_offset_in_pixels()
		w, h = self.size
		src_rect = Rect(x0, y0, w, h)
		img = self.surface_image.subsurface(src_rect)
		surf.blit(img, (0, 0))
	
	def draw_tiles(self, surf, world):
		cells = world.terrain
		x0, y0 = self.scroll_offset_in_pixels()
		i0, j0 = self.topleft_cell_coords()
		wisize, wjsize = world.size
		visize, vjsize = self.size_in_cells()
		i1 = min(i0 + visize, wisize)
		j1 = min(j0 + vjsize, wjsize)
		dx = TILE_WIDTH
		dy = TILE_HEIGHT
		for i in xrange(i0, i1):
			for j in xrange(j0, j1):
				cells[i, j].draw(surf, i * dx - x0, j * dy - y0)
	
	def draw_units(self, surf, world):
		units = sorted(world.units, key = lambda u: u.position[1])
		selected = world.selected_units
		for unit in units:
			p = self.world_to_screen(unit.position)
			r = self.screen_rect_for_unit(unit)
			if unit in selected:
				self.draw_selection_ring(surf, p, unit.radius)
			if unit.is_alive:
				#surf.blit(unit.image, r)
				unit.draw(surf, r)
				self.draw_unit_health_bar(surf, unit, r)
				self.draw_unit_build_progress(surf, unit, r)
				if unit.under_construction:
					self.draw_construction_barriers(surf, unit, r)
			elif unit.leaves_corpse:
				self.draw_unit_corpse(surf, unit, r)

	def draw_selection_ring(self, surf, p, radius):
		r0 = selection_ring.get_rect()
		width = int(round(3 * radius * TILE_WIDTH))
		scale = width / r0.width
		img = rotozoom(selection_ring, 0, scale)
		r1 = img.get_rect()
		r1.center = p
		surf.blit(img, r1)

	def draw_unit_health_bar(self, surf, unit, unit_rect):
		h = unit.health / unit.max_health
		if h < 1.0:
			full_width = unit_rect.width
			health_width = int(full_width * h)
			r = Rect(0, 0, full_width, 3)
			r.midbottom = unit_rect.midtop
			surf.fill((0, 0, 0), r)
			r.width = health_width
			surf.fill((0, 255, 0), r)
	
	def draw_unit_build_progress(self, surf, unit, unit_rect):
		p = unit.build_progress_fraction
		if p is not None:
			full_width = unit_rect.width
			width = int(full_width * p)
			r = Rect(0, 0, full_width, 3)
			r.midbottom = (unit_rect.centerx, unit_rect.bottom - 5)
			rb = r.inflate(4, 4)
			surf.fill((0, 0, 0), rb)
			r.width = width
			surf.fill((255, 255, 255), r)

	def draw_construction_barriers(self, surf, unit, unit_rect):
		image = construction_image
		r = image.get_rect()
		r.midbottom = unit_rect.midbottom
		surf.blit(image, r)

	def draw_unit_corpse(self, surf, unit, unit_rect):
		image = rotate(unit.image, 90)
		r = image.get_rect()
		r.midbottom = unit_rect.midbottom
		surf.blit(image, r)
	
	def draw_site_selection_feedback(self, surf, screen_pt, unit_class):
		site_size = unit_class.site_size_required()
		world_pt = self.screen_to_world(screen_pt)
		world_min = rint(world_pt - 0.5 * site_size)
		building_site = self.world.site_suitable_for_unit_class(world_pt, unit_class)
		if building_site is not None:
			color = (0, 255, 0)
		else:
			color = (255, 0, 0)
		rect = Rect((0, 0), site_size * TILE_SIZE)
		rect.topleft = self.world_to_screen(world_min)
		frame_rect(surf, color, rect)

	def mouse_move(self, e):
		self.client.world_mouse_move(e)
	
	def mouse_down(self, e):
		e.dict['double'] = e.time - self.last_click_time <= 250
		self.last_click_time = e.time
		self.client.world_click(e)
	
	def perform_unit_selection(self, e):
		world = self.world
		selection = world.selected_units
		if not e.shift or e.double:
			world.clear_selected_units()
		unit = self.find_unit_at_screen_point(e.local)
		if unit and unit.is_alive:
			#if unit.team_no == 0 or options.enemy_control_option:
				if e.double:
					if e.shift:
						range = None
					else:
						range = 2.0
					self.select_nearby_units_of_type(unit.position, unit.__class__, range)
				else:
					if unit not in selection:
						world.select_unit(unit)
					else:
						world.deselect_unit(unit)

	def select_nearby_units_of_type(self, p, cls, range):
		world = self.world
		selection = world.selected_units
		for unit in world.units:
			#if unit.team_no == 0:
				if unit.__class__ is cls \
					and (range is None or mag(p - unit.position2) <= range):
						selection.add(unit)
	
	def find_unit_at_screen_point(self, p):
		unit_found = None
		for unit in self.world.units:
			r = self.display_rect_for_unit(unit)
			if r.collidepoint(p):
				ranking = unit.selection_ranking()
				if not unit_found or unit.selection_ranking() > ranking_found:
					unit_found = unit
					ranking_found = ranking
		return unit_found

	def screen_rect_for_unit(self, unit):
		cx, cy = self.world_to_screen(unit.position)
		#r = unit.image.get_rect()
		r = unit.get_image_rect()
		r.centerx = cx
		r.bottom = cy + r.width // 2 * TILE_HEIGHT // TILE_WIDTH
		return r

	def display_rect_for_unit(self, unit):
		r = self.screen_rect_for_unit(unit)
		if not unit.is_alive and unit.leaves_corpse:
			p = r.midbottom
			r = Rect(0, 0, r.height, r.width)
			r.midbottom = p
		return r

